(      Title:  Parsing the Input Stream Across Lines or Blocks
     Library:  Parsing ANS Forth Strings and the Input Stream
        File:  linstream.fs
   Test file:  parsing-test.fs
    Log File:  parsing.log
      Author:  David N. Williams
     License:  LGPL
     Version:  1.0.3
     Revised:  February 14, 2012

Some of this file is derived from Wil Baden's Tool Belt, which
is in the public domain.  It also contains words from various
other sources, which we believe to be compatible with the LGPL. 
For the sake of the LGPL, the rest is
)
\ Copyright (C) 1996-2012 David N. Williams
(
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or at your option any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

If you take advantage of the option in the LGPL to put a
particular version of this library under the GPL, the author[s]
would regard it as polite if you would put any direct
modifications under the LGPL as well, and include a copy of this
request near the beginning of the modified library source.  A
"direct modification" is one that enhances or extends the
library in line with its original concept, as opposed to
developing a distinct application or library which might use it.

This file is mostly ANS Forth compatible up to case dependence
and [UNDEFINED].

It is third in order of dependency of three modules that
comprise a library for parsing Forth strings, the parse area,
and the input stream across lines or blocks:

  fstrings
  parse-area
  linstream

See the file fstrings.fs for the notation and terminology used
by all three modules.

This module is for parsing the input stream across lines or
blocks.  Sources include Wil Baden and Bruce McFarling.

The words in this library ignore any counted or measured string
model for storage in memory.

See the file parsing-ext.c for a C implementation as a pfe
external module, and the file parsing-test.fs for Hayes-style
tests based on Anton Ertl's ttester.fs.

Since they are based on REFILL, words that act across lines are
expected to work across blocks as well, but have not been tested
with blocks.

The code is intended to be character clean.
)
decimal

REQUIRE fstrings.fs
REQUIRE parse-area.fs

MODULE: linstream
  ALSO fstrings
  ALSO parse-area
  ' fstrings >private >order
BEGIN-MODULE

\ *** PUBLIC INPUT STREAM WORDS
(
  \\  ?EMIT-CR  NEXT-INSTREAM-NAME
  S-SEEK-INSTREAM  |S|-SEEK-INSTREAM  (*

    Not implemented:
    SKIP-INSTREAM  SCAN-INSTREAM  SEEK-INSTREAM
    {}SKIP-INSTREAM  {}SCAN-INSTREAM  {}SEEK-INSTREAM

    Uncommentable:
    ((  ---
)

PUBLIC:

\ *** INPUT STREAM ACROSS LINES OR BLOCKS
(
With the exception of \\, all across-line words in this section
echo CR's when the lines are being entered interactively at the
terminal.  That feature has been tested by copying and pasting
the file parsing-test.fs into a terminal window in which pfe or
gforth is running.  See parsing-test.fs for an explanation.
)

: \\  ( "<eof>" -- )				\ 2002
(
Since REFILL does nothing but return false when the source is a
string, the following covers the string case, too.  It should
not be used when the input source is the terminal, as then it
causes an infinite loop.

We learned to recognize the equivalence of the phrase "-1 PARSE
2DROP" to EMPTY-PARSE-AREA [inside the loop in the Tool Belt
version] from Bruce McFarling, Factoring and Visibility [redux],
comp.lang.forth, December 4, 2006.
)
  empty-parse-area BEGIN refill 0= UNTIL ;

: ?emit-cr  ( -- )
(
Emit a CR if the input source is the terminal.
)
  source-id 0= IF cr THEN ;

: next-instream-name  ( "<name>" -- word.s )  \ 2002 Next-Word
(
Leave the next word in the input stream, and advance the input
stream to just after the first of any trailing whitespace
characters.  If word.s is empty, at most whitespace was found. 
A parsing implementation of Wil Baden's Next-Word, plus
conditional echoing of CR's.
)
  BEGIN parse-name dup IF EXIT THEN ?emit-cr refill
  WHILE 2drop REPEAT ;

(
The behavior of S-SEEK-INSTREAM and |S|-SEEK-INSTREAM for empty
string input is not specified below, but should be derivable
from that of SEARCH.  We have not thought through what might be
a desirable behavior.
)

: s-seek-instream  ( "<pat>" pat.s -- flag )
(
Assume pat.s to be nonempty.  Advance the input stream to just
after the next occurrence of the string, matching embedded
whitespace characters exactly, and leave true if found.  If the
string is not found after REFILL attempts, empty the parse area
and leave false.
)
  ( s) 2>r
  BEGIN
    parse-area@ 2r@ ( pa.s s)
    search ( pa.s false | &found #rem true)
  0= WHILE  \ not found
    ( pa.s) 2drop ?emit-cr
    refill 0= IF  \ end of file
      2rdrop empty-parse-area false EXIT THEN
  REPEAT  \ found
  r> /string parse-area! rdrop true ;

: |s|-seek-instream  ( "<pat>" pat.s -- flag )
(
Assume pat.s to be nonempty.  Advance the input stream to just
after the next occurrence of the string, matching embedded
whitespace characters exactly, delimited by whitespace on both
ends; and leave true if found. If the delimited string is not
found after REFILL attempts, empty the parse area and leave
false.
)
  ( s) 2>r
  BEGIN
    2r@ s-seek-instream
  WHILE
    parse-area@ drop r@ chars- r@ source white-delimited?
    IF 2rdrop true EXIT THEN
    >in++  \ jog for next seek
  REPEAT 2rdrop false ;

: (*  ( -- )
(
Advance the input stream to just after next occurrence of the
string consisting of the asterisk and right parenthesis
characters, delimited by whitespace.
)
  s" *)" |s|-seek-instream
  0= ABORT" ***Missing *) terminator"; immediate

(*
The point of making comment terminators whitespace delimited is
explained by Stephen Pelc in comp.lang.forth, "Multi-line
comment markers", August 9, 2005:
   
  It then makes sense to require the comment block end marker
  also to be whitespace delimited to prevent problems with Forth
  word names in the text.
*)

\ : ((  ( -- )  s" ))"  |s|-seek-instream drop ; immediate
\ : --- ( -- )  s" ---" |s|-seek-instream drop ; immediate
(*
Advance the input stream to just after next occurrence of the
whitespace delimited string "))", respectively, "---".
*)

END-MODULE
