(      Title:  Parsing the Parse Area
     Library:  Parsing ANS Forth Strings and the Input Stream
        File:  parse-area.fs
   Test file:  parsing-test.fs
    Log File:  parsing.log
      Author:  David N. Williams
     License:  LGPL
     Version:  1.0.3
     Revised:  February 14, 2012

Besides work by the author, this file contains words from
various sources, which we believe to be compatible with the
LGPL.  For the sake of the LGPL, the part due to the author is
)
\ Copyright (C) 1996-2012 David N. Williams
(
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or at your option any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

If you take advantage of the option in the LGPL to put a
particular version of this library under the GPL, the author[s]
would regard it as polite if you would put any direct
modifications under the LGPL as well, and include a copy of this
request near the beginning of the modified library source.  A
"direct modification" is one that enhances or extends the
library in line with its original concept, as opposed to
developing a distinct application or library which might use it.

This file is mostly ANS Forth compatible up to case dependence.

It is the second in order of dependency of three modules that
comprise a library for parsing Forth strings, the parse area,
and the input stream across lines or blocks:

  fstrings
  parse-area
  linstream

See the file fstrings.fs for the notation and terminology used
by all three modules.

This module is for parsing the parse area of the input stream.
Sources include Michael Gassanenko, Bernd Paysan, Frederick
Warren, and Leo Wong.

The words in this library ignore any counted or measured string
model for storage in memory.


See the file parsing-ext.c for a C implementation of the parsing
library collection as a pfe external module, and the file
parsing-test.fs for Hayes-style tests of the parsing library
based on Anton Ertl's ttester.fs.

The code is intended to be character clean.
)
decimal

REQUIRE fstrings.fs

MODULE: parse-area
  ALSO fstrings
  ' fstrings >private >order
BEGIN-MODULE

\ *** PUBLIC PARSE AREA WORDS
(
  PARSE-AREA@  PARSE-AREA!  EMPTY-PARSE-AREA  PARSE-AREA-EMPTY?
  PARSE-NAME  PREPARSE-NAME  PARSE-NAME-AWAY  {}PARSE
  >IN++  >IN--

    Defined in parsing-test.fs:
    >IN+  >IN-  

    Not implemented:
    SPACE|PUNCT-PARSE
    S-PARSE  PARSE-AWAY  {}PARSE-AWAY
)

PUBLIC:

\ *** PARSE AREA

0 [IF]
The words PARSE-AREA@ and PARSE-AREA! provide the basic
interface for operating on the parse area with fstring words.
For example:

: parse-area-skip  ( char -- )
  ( char) >r parse-area@ r> skip parse-area! ;

It is left to the user to define such words as needed.
[THEN]

: parse-area@  ( -- unparsed.s )
(
Leave the parse area portion of the input buffer as a string.
Factored from Bernd Paysan's version of Frederick Warren's $>,
comp.lang.forth, March, 2000.  Named by Michael Gassanenko.
)
  source ( &inbuf #inbuf) >in @ ( #parsed) /string ;

: parse-area!  ( unparsed.s -- )
(
The unparsed.s string is assumed to lie at the end of the input
buffer.  Advance the input stream so that unparsed.s is the
parse area.  Named by Michael Gassanenko.
)
  ( addr len) source nip ( #tib) swap - >in ! ( addr) drop ;

: empty-parse-area  ( -- )
(
Exhaust the input buffer.

Note:  REFILL can't always be used for this.  When the source is
a string, REFILL does nothing.  When it is a file, one may not
want the current input buffer to be written over.
)
  source ( addr #chars) >in ! ( addr) drop ;

: parse-area-empty?  ( -- flag )  parse-area@ nip 0= ;

[UNDEFINED] PARSE-NAME [IF]
: parse-name  ( "<name>" -- word.s )             \ CORE EXT
(
Leave the next word in the parse area, and advance the input
stream to just after the first of any trailing whitespace
characters.  If there is no word in the parse area, word.s is
the right empty string, and the parse area is left exhausted.

NOTE:  Forth 200x does not require that the empty string be the
right empty string.  The native version of PARSE-NAME does
produce the right empty string for gforth and pfe, but not
always for iForth.
)
  parse-area@ separate-word 2swap parse-area! ;
[THEN]

: preparse-name  ( "<name>" -- word.s )
(
Get the next word in the parse area without advancing the input
stream.  Leo Wong's PREPARSE.
)
  >in @ >r parse-name r> >in ! ;

: parse-name-away  ( "<name>" -- )
(
When the parse area is not all white space, parse away the next
word and the first of any trailing whitespace characters.
Otherwise empty the parse area.  A logical companion to
PREPARSE-NAME.
) 
  parse-name 2drop ;

: {}parse  ( "<char>" &chars -- parsed.s char|0 )
(
Scan the parse area for any of the null-terminated sequence of
characters at &chars.  If a delimiting character is found, leave
it as char.  If a delimiting character is not found, leave zero
instead.

If a delimiting character is found, the parse area is left
positioned just after it, and parsed.s represents the former
part of the parse area up to just before char.  If a delimiting
character is not found, the input stream is left exhausted, and
parsed.s represents the former parse area.
)
  >r parse-area@ r> {}separate ( char) >r
  r@ IF 2swap parse-area! ELSE empty-parse-area THEN r> ;
 
: >in++  ( -- )
(
Step the parse area one character forward, unless it is
exhausted.
)
  source nip >in @ 1+ min >in ! ;

: >in--  ( -- )
(
Step the parse area one character back, unless it is positioned
at the start of the input buffer.
)
  >in @ 1- 0 max >in ! ;

END-MODULE
